/* Copyright 2008 BORDEAUX I UNIVERSITY & INRIA 
**
** This file is part of the PaStiX parallel sparse matrix package.
**
** This software is governed by the CeCILL-C license under French law
** and abiding by the rules of distribution of free software. You can
** use, modify and/or redistribute the software under the terms of the
** CeCILL-C license as circulated by CEA, CNRS and INRIA at the following
** URL: "http://www.cecill.info".
** 
** As a counterpart to the access to the source code and rights to copy,
** modify and redistribute granted by the license, users are provided
** only with a limited warranty and the software's author, the holder of
** the economic rights, and the successive licensors have only limited
** liability.
** 
** In this respect, the user's attention is drawn to the risks associated
** with loading, using, modifying and/or developing or reproducing the
** software by the user in light of its specific status of free software,
** that may mean that it is complicated to manipulate, and that also
** therefore means that it is reserved for developers and experienced
** professionals having in-depth computer knowledge. Users are therefore
** encouraged to load and test the software's suitability as regards
** their requirements in conditions enabling the security of their
** systems and/or data to be ensured and, more generally, to use and
** operate it in the same conditions as regards security.
** 
** The fact that you are presently reading this means that you have had
** knowledge of the CeCILL-C license and that you accept its terms.
*/
/************************************************************/
/**                                                        **/
/**   NAME       : queue.h                                 **/
/**                                                        **/
/**   AUTHORS    : Pascal HENON                            **/
/**                                                        **/
/**   FUNCTION   : queue of PASTIX_INT that sorts elements        **/
/**                in ascending way according to a         **/
/**                PASTIX_FLOAT key                               **/
/**   DATES      : # Version 0.0  : from : 22 jul 1998     **/
/**                                 to     08 sep 1998     **/
/**                                                        **/
/************************************************************/

#ifndef QUEUE_H
#define QUEUE_H
#define pastix_int_t PASTIX_INT
#define PASTIX_SUCCESS NO_ERR

typedef struct pastix_queue_item_s {
    double          key1;               /*+ Key 1 of the element             +*/
    double          key2;               /*+ Key 2 of the element             +*/
    /* void           *eltptr;             /\*+ Pointer to the element           +*\/ */
    pastix_int_t    eltptr;                 /*+ Pointer to the element           +*/
} pastix_queue_item_t;

typedef struct pastix_queue_s {
    pastix_int_t         size;          /*+ Allocated memory size             +*/
    pastix_int_t         used;          /*+ Number of element in the queue    +*/
    pastix_queue_item_t *elttab;        /*+ Array of the element              +*/
} pastix_queue_t;

int          pqueueInit( pastix_queue_t *, pastix_int_t);
void         pqueueExit( pastix_queue_t *);
pastix_int_t pqueueSize( pastix_queue_t *);
void         pqueueClear(pastix_queue_t *);
/*void         pqueueCopy(pastix_queue_t *, pastix_queue_t *);*/
void         pqueuePush2(pastix_queue_t *, pastix_int_t, double, double);
pastix_int_t pqueueRead (pastix_queue_t *);
pastix_int_t pqueuePop2 (pastix_queue_t *, double *, double *);

static inline void
pqueuePush1(pastix_queue_t *q, pastix_int_t elt, double key1) {
    return pqueuePush2( q, elt, key1, 0 );
}

static inline pastix_int_t
pqueuePop(pastix_queue_t *q){
    return pqueuePop2(q, NULL, NULL);
}

static inline pastix_int_t
pqueuePop1(pastix_queue_t *q, double *key1){
    return pqueuePop2(q, key1, NULL);
}

/* int     queuePossess    (Queue *, pastix_int_t); */
/* void    queuePrint      (Queue *); */

/* static pastix_int_t compWith2keys(Queue *, pastix_int_t, pastix_int_t); */

/*
**  The type and structure definitions.
*/

typedef struct Queue_ {
  PASTIX_INT        size;                  /*+ Allocated memory size             +*/ 
  PASTIX_INT        used;                  /*+ Number of element in the queue    +*/
  PASTIX_INT    *   elttab;                /*+ Array of the element              +*/
  double *   keytab;                /*+ Array of keys                     +*/
  PASTIX_INT    *   keytab2;               /*+ Another array of keys             +*/
} Queue;


#define static

int     queueInit       (Queue *, PASTIX_INT size);
void    queueExit       (Queue *);
Queue * queueCopy       (Queue *dst, Queue *src);
void    queueAdd        (Queue *, PASTIX_INT, double);
void    queueAdd2       (Queue *, PASTIX_INT, double, PASTIX_INT);
PASTIX_INT     queueGet        (Queue *);
PASTIX_INT     queueSize       (Queue *);
void    queueClear      (Queue *);
PASTIX_INT     queueRead       (Queue *);
PASTIX_INT     queueGet2       (Queue *, double *, PASTIX_INT *);
int     queuePossess    (Queue *, PASTIX_INT);
void    queuePrint      (Queue *);

static PASTIX_INT compWith2keys(Queue *, PASTIX_INT, PASTIX_INT);
#undef static
#endif
