/* Copyright 2008 BORDEAUX I UNIVERSITY & INRIA 
**
** This file is part of the PaStiX parallel sparse matrix package.
**
** This software is governed by the CeCILL-C license under French law
** and abiding by the rules of distribution of free software. You can
** use, modify and/or redistribute the software under the terms of the
** CeCILL-C license as circulated by CEA, CNRS and INRIA at the following
** URL: "http://www.cecill.info".
** 
** As a counterpart to the access to the source code and rights to copy,
** modify and redistribute granted by the license, users are provided
** only with a limited warranty and the software's author, the holder of
** the economic rights, and the successive licensors have only limited
** liability.
** 
** In this respect, the user's attention is drawn to the risks associated
** with loading, using, modifying and/or developing or reproducing the
** software by the user in light of its specific status of free software,
** that may mean that it is complicated to manipulate, and that also
** therefore means that it is reserved for developers and experienced
** professionals having in-depth computer knowledge. Users are therefore
** encouraged to load and test the software's suitability as regards
** their requirements in conditions enabling the security of their
** systems and/or data to be ensured and, more generally, to use and
** operate it in the same conditions as regards security.
** 
** The fact that you are presently reading this means that you have had
** knowledge of the CeCILL-C license and that you accept its terms.
*/
/**
 * @file starpu_zkernels.h
 * @precisions normal z -> s d c
 */
#ifndef STARPU_ZKERNELS_H
#define STARPU_ZKERNELS_H

#define MAT_zaxpy( m, n, alpha,                   \
                   A, lda,                        \
                   B, ldb )                       \
  {                                               \
        pastix_int_t i,j;                                 \
    for (j=0; j<n; j++) {                         \
      for (i=0; i<m; i++) {                       \
        B[j*ldb+i] = B[j*ldb+i] - A[j*lda+i];     \
      }                                           \
    }                                             \
  }

#define MAT_zaxpyt( m, n, alpha,                  \
                    A, lda,                       \
                    B, ldb )                      \
  {                                               \
        pastix_int_t i,j;                                 \
    for (j=0; j<n; j++) {                         \
      for (i=0; i<m; i++) {                       \
        B[j*ldb+i] = B[j*ldb+i] - A[j+i*lda];     \
      }                                           \
    }                                             \
}

#  ifndef STARPU_PROF_CALLBACK_DEFINE
#  define STARPU_PROF_CALLBACK_DEFINE
void starpu_prof_callback(void *callback_arg);
#  endif

/* For old kernels ===> TRASH when not used anymore*/
#define ARCH_CPU  0
#define ARCH_CUDA 1


#define trsm_starpu_common                API_CALL(trsm_starpu_common)
#define xxtrf_starpu_common               API_CALL(xxtrf_starpu_common)
#define trfsp1d_starpu_common             API_CALL(trfsp1d_starpu_common)
#define trfsp1d_gemm_starpu_common        API_CALL(trfsp1d_gemm_starpu_common)
#define trfsp1d_sparse_gemm_starpu_common API_CALL(trfsp1d_sparse_gemm_starpu_common)
#define trsm_starpu_cpu                   API_CALL(trsm_starpu_cpu)
#define xxtrf_starpu_cpu                  API_CALL(xxtrf_starpu_cpu)
#define trfsp1d_starpu_cpu                API_CALL(trfsp1d_starpu_cpu)
#define trfsp1d_gemm_starpu_cpu           API_CALL(trfsp1d_gemm_starpu_cpu)
#define trfsp1d_sparse_gemm_starpu_cpu    API_CALL(trfsp1d_sparse_gemm_starpu_cpu)
#define trsm_starpu_cuda                  API_CALL(trsm_starpu_cuda)
#define xxtrf_starpu_cuda                 API_CALL(xxtrf_starpu_cuda)
#define trfsp1d_starpu_cuda               API_CALL(trfsp1d_starpu_cuda)
#define trfsp1d_gemm_starpu_cuda          API_CALL(trfsp1d_gemm_starpu_cuda)
#define trfsp1d_sparse_gemm_starpu_cuda   API_CALL(trfsp1d_sparse_gemm_starpu_cuda)

void trfsp1d_starpu_cpu(void * buffers[], void * _args);
void xxtrf_starpu_cpu(void * buffers[], void * _args);
void trsm_starpu_cpu(void * buffers[], void * _args);
void trfsp1d_gemm_starpu_cpu(void * buffers[], void * _args);
void trfsp1d_sparse_gemm_starpu_cpu(void * buffers[], void * _args);
void trfsp1d_starpu_cuda(void * buffers[], void * _args);
void xxtrf_starpu_cuda(void * buffers[], void * _args);
void trsm_starpu_cuda(void * buffers[], void * _args);
void trfsp1d_gemm_starpu_cuda(void * buffers[], void * _args);
void trfsp1d_sparse_gemm_starpu_cuda(void * buffers[], void * _args);

#define init_coeftab_cpu API_CALL(init_coeftab_cpu)
#define init_coeftab_cuda API_CALL(init_coeftab_cuda)
void init_coeftab_cpu(void * buffers[], void * _args);
void init_coeftab_cuda(void * buffers[], void * _args);

#define fill_coeftab_cpu API_CALL(fill_coeftab_cpu)
#define fill_coeftab_cuda API_CALL(fill_coeftab_cuda)
void fill_coeftab_cpu(void * buffers[], void * _args);
void fill_coeftab_cuda(void * buffers[], void * _args);

#define empty_task_func API_CALL(empty_task_func)
void empty_task_func(void * buffers[], void * _args);

#endif
